/**
 * Copyright (c) 2012 - 2019 Data In Motion and others.
 * All rights reserved. 
 * 
 * This program and the accompanying materials are made available under the terms of the 
 * Eclipse Public License v1.0 which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * Contributors:
 *     Data In Motion - initial API and implementation
 */
package org.gecko.rsa.discovery.ma.handler;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.OutputStream;
import java.nio.ByteBuffer;
import java.util.Collection;
import java.util.Map;
import java.util.Optional;
import java.util.concurrent.ConcurrentHashMap;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.gecko.osgi.messaging.Message;
import org.gecko.osgi.messaging.MessagingService;
import org.gecko.rsa.api.ExportEndpointHandler;
import org.gecko.rsa.core.EndPointDeSerializer;
import org.gecko.rsa.core.converter.EndpointDescriptionConverter;
import org.gecko.rsa.discovery.EndpointDiscovery;
import org.gecko.rsa.discovery.ImportingEndpointManager;
import org.gecko.rsa.model.rsa.EndpointDescriptions;
import org.gecko.rsa.model.rsa.Property;
import org.gecko.rsa.model.rsa.RSAFactory;
import org.osgi.framework.BundleContext;
import org.osgi.framework.Constants;
import org.osgi.service.remoteserviceadmin.EndpointDescription;
import org.osgi.service.remoteserviceadmin.RemoteConstants;
import org.osgi.util.promise.Promise;
import org.osgi.util.pushstream.PushStream;

/**
 * 
 * @author mark
 * @since 05.04.2019
 */
public class MessageAdapterDiscoverHandler implements ExportEndpointHandler {

	private static final Logger logger = Logger.getLogger(MessageAdapterDiscoverHandler.class.getName());
	private final Map<String, org.gecko.rsa.model.rsa.EndpointDescription> importingDescriptions = new ConcurrentHashMap<String, org.gecko.rsa.model.rsa.EndpointDescription>();
	private final Map<String, org.gecko.rsa.model.rsa.EndpointDescription> exportingDescriptions = new ConcurrentHashMap<String, org.gecko.rsa.model.rsa.EndpointDescription>();
	private final EndpointDescriptionConverter converter = new EndpointDescriptionConverter();
	private final EndPointDeSerializer deSerializer = new EndPointDeSerializer();
	private final EndpointDiscovery discovery;
	private final MessagingService messaging;
	private final String frameworkUUID;

	private PushStream<Message> importSubscribe;
	private PushStream<Message> exportSubscribe;
	private PushStream<Message> removeSubscribe;
	private PushStream<Message> modifySubscribe;
	private PushStream<Message> announceSubscribe;

	/**
	 * Creates a new instance.
	 */
	public MessageAdapterDiscoverHandler(MessagingService messaging, BundleContext ctx, MADiscoveryConfig config) {
		this.messaging = messaging;
		this.discovery = new EndpointDiscovery(ctx, config.name(), this);
		frameworkUUID = ctx.getProperty(Constants.FRAMEWORK_UUID);
	}

	public void start() {
		if (discovery == null || messaging == null) {
			throw new IllegalStateException("There is a missing discovery or messaging to start this handler");
		}
		discovery.start();
		initializeSubscriptions();
		callForDescription();
	}

	public void stop() {
		if (discovery != null && discovery.isRunning()) {
			discovery.stop();
		}
		if (importSubscribe != null) {
			importSubscribe.close();
		}
		if (modifySubscribe != null) {
			modifySubscribe.close();
		}
		if (removeSubscribe != null) {
			removeSubscribe.close();
		}
	}

	/**
	 * Initializes all subscriptions
	 */
	private void initializeSubscriptions() {
		try {
			exportSubscribe = messaging.subscribe(MADiscoveryConstants.TOPIC_ADD_SERVICE);
			exportSubscribe.forEach(this::notifyAdd);
		} catch (Exception e) {
			logger.log(Level.SEVERE, String.format("Error subscribing for topic ", MADiscoveryConstants.TOPIC_ADD_SERVICE), e);
		}
		try {
			removeSubscribe = messaging.subscribe(MADiscoveryConstants.TOPIC_REMOVE_SERVICE);
			removeSubscribe.forEach(this::notifyRemove);
		} catch (Exception e) {
			logger.log(Level.SEVERE, String.format("Error subscribing for topic ", MADiscoveryConstants.TOPIC_REMOVE_SERVICE), e);
		}
		try {
			modifySubscribe = messaging.subscribe(MADiscoveryConstants.TOPIC_MODIFY_SERVICE);
			modifySubscribe.forEach(this::notifyModify);
		} catch (Exception e) {
			logger.log(Level.SEVERE, String.format("Error subscribing for topic ", MADiscoveryConstants.TOPIC_MODIFY_SERVICE), e);
		}
		try {
			announceSubscribe = messaging.subscribe(String.format(MADiscoveryConstants.TOPIC_ANNOUNCED_SERVICE, frameworkUUID));
			announceSubscribe.forEach(this::notifyModify);
		} catch (Exception e) {
			logger.log(Level.SEVERE, String.format("Error subscribing for topic ", MADiscoveryConstants.TOPIC_MODIFY_SERVICE), e);
		}
		try {
			importSubscribe = messaging.subscribe(MADiscoveryConstants.TOPIC_GET_DESCRIPTION);
			importSubscribe.forEach(this::publishAllDescriptions);
		} catch (Exception e) {
			logger.log(Level.SEVERE, String.format("Error subscribing for topic ", MADiscoveryConstants.TOPIC_GET_DESCRIPTION), e);
		}
	}

	/* 
	 * (non-Javadoc)
	 * @see org.gecko.rsa.api.ExportEndpointHandler#exportServiceDescription(org.osgi.service.remoteserviceadmin.EndpointDescription)
	 */
	@Override
	public void exportServiceDescription(EndpointDescription endpoint) throws Exception {
		org.gecko.rsa.model.rsa.EndpointDescription ed = converter.caseOSGiEndpointDescription(endpoint);
		String key = getKey(ed);
		if (key != null) {
			exportingDescriptions.put(key, ed);
			publishEndpointDescription(ed, MADiscoveryConstants.TOPIC_ADD_SERVICE);
			logger.log(Level.INFO, String.format("Exporting new endpoint via message adapter. Endpoint: %s, Key: %s", endpoint, key));
		} else {
			logger.severe(String.format("Exporting new endpoint failed because of missing endpoint.id. Endpoint: %s, Key: %s" + ed, endpoint, key));
		}
	}

	/* 
	 * (non-Javadoc)
	 * @see org.gecko.rsa.api.ExportEndpointHandler#modifyServiceDescription(org.osgi.service.remoteserviceadmin.EndpointDescription)
	 */
	@Override
	public void modifyServiceDescription(EndpointDescription endpoint) throws Exception {
		org.gecko.rsa.model.rsa.EndpointDescription ed = converter.caseOSGiEndpointDescription(endpoint);
		String key = getKey(ed);
		if (key != null) {
			if (exportingDescriptions.put(key, ed) != null) {
				publishEndpointDescription(ed, MADiscoveryConstants.TOPIC_MODIFY_SERVICE);
				logger.log(Level.INFO, String.format("Exporting changed endpoint via message adapter. Endpoint: %s, Key: %s", endpoint, key));
			} else {
				exportServiceDescription(endpoint);
			}
		} else {
			logger.severe(String.format("Exporting changed endpoint failed because of missing endpoint.id. Endpoint: %s, Key: %s" + ed, endpoint, key));
		}
	}

	/* 
	 * (non-Javadoc)
	 * @see org.gecko.rsa.api.ExportEndpointHandler#removeServiceDescription(org.osgi.service.remoteserviceadmin.EndpointDescription)
	 */
	@Override
	public void removeServiceDescription(EndpointDescription endpoint) throws Exception {
		org.gecko.rsa.model.rsa.EndpointDescription ed = converter.caseOSGiEndpointDescription(endpoint);
		String key = getKey(ed);
		if (key != null) {
			if (exportingDescriptions.remove(key) != null) {
				publishEndpointDescription(ed, MADiscoveryConstants.TOPIC_REMOVE_SERVICE);
				logger.log(Level.INFO, String.format("Exporting removed endpoint via message adapter. Endpoint: %s, Key: %s", endpoint, key));
			} 
		} else {
			logger.severe(String.format("Exporting removed endpoint failed because of missing endpoint.id. Endpoint: %s, Key: %s" + ed, endpoint, key));
		}
	}

	/**
	 * Returns the key, which is the end-point id
	 * @param endpointDescription the endpoint description
	 * @return the endpoint id property or null
	 */
	private String getKey(org.gecko.rsa.model.rsa.EndpointDescription endpointDescription) {
		Optional<Property> endpointId = endpointDescription.getProperty().stream().filter(p->p.getName().equals(RemoteConstants.ENDPOINT_ID)).findFirst();
		return endpointId.map(p->p.getValue()).orElse(null);
	}

	/**
	 * Received a add message and notifies all listeners about this end-point change
	 * @param message the received message
	 */
	private void notifyAdd(Message message) {
		ByteArrayInputStream bais = new ByteArrayInputStream(message.payload().array());
		ImportingEndpointManager manager = discovery.getImportingEndpointManager();
		Promise<EndpointDescriptions> endpointDescriptionsProm = deSerializer.deserialize(bais);
		endpointDescriptionsProm.thenAccept(d->{
			synchronized (d) {
				for (org.gecko.rsa.model.rsa.EndpointDescription ed : d.getEndpointDescription()) {
					String key = getKey(ed);
					if (importingDescriptions.containsKey(key) || exportingDescriptions.containsKey(key)) {
						logger.info("Detected and new incoming endpoint, that already exists. Ignore it");
						continue;
					} 
					importingDescriptions.put(key, ed);
					manager.addImported(ed);
					logger.fine("Detected and incoming endpoint andsent 'ADD' event");
				}
			}
		}).onFailure(t->logger.log(Level.SEVERE, "Detected an error for an incoming endpoint. Cannot send ADD event", t));
	}

	/**
	 * Received a modification message and notifies all listeners about this end-point change.
	 * If there is no end-point description, it will be created and an add event will be sent. 
	 * Otherwise a modified event will be sent
	 * @param message the received message
	 */
	private void notifyModify(Message message) {
		ByteArrayInputStream bais = new ByteArrayInputStream(message.payload().array());
		ImportingEndpointManager manager = discovery.getImportingEndpointManager();
		Promise<EndpointDescriptions> endpointDescriptionsProm = deSerializer.deserialize(bais);
		endpointDescriptionsProm.thenAccept(d->{
			synchronized (d) {
				for (org.gecko.rsa.model.rsa.EndpointDescription ed : d.getEndpointDescription()) {
					String key = getKey(ed);
					if (exportingDescriptions.containsKey(key)) {
						logger.info("Detected an incoming modified endpoint, that already exists. Ignore it");
						continue;
					}
					org.gecko.rsa.model.rsa.EndpointDescription edModify = importingDescriptions.put(key, ed);
					if (edModify != null) {
						manager.modifyImported(edModify);
						logger.info(String.format("Detected and incoming endpoint sent '%s' event", "MODIFY"));
					} else {
						manager.addImported(ed);
					}
				}
			}
		}).onFailure(t->logger.log(Level.SEVERE, "Detected an error for a modified endpoint. Cannot send MODIFY or ADD event", t));
	}

	/**
	 * Received a remove message and notifies all listeners about this end-point change.
	 * @param message the received message
	 */
	private void notifyRemove(Message message) {
		ByteArrayInputStream bais = new ByteArrayInputStream(message.payload().array());
		ImportingEndpointManager manager = discovery.getImportingEndpointManager();
		Promise<EndpointDescriptions> endpointDescriptionsProm = deSerializer.deserialize(bais);
		endpointDescriptionsProm.thenAccept((d)->{
			synchronized (d) {
				for (org.gecko.rsa.model.rsa.EndpointDescription ed : d.getEndpointDescription()) {
					String key = getKey(ed);
					org.gecko.rsa.model.rsa.EndpointDescription edRemove = importingDescriptions.remove(key);
					if (edRemove != null) {
						manager.removeImported(edRemove);
						logger.info("Detected and incoming endpoint sent REMOVE event");
					} else {
						logger.info("No endpoint was found, that could be removed");
					}
				}
			}
		}).onFailure(t->logger.log(Level.SEVERE, "Detected an error for a remove endpoint. Cannot send REOMVE event", t));
	}

	/**
	 * Publishes a end-point description holder
	 * @param descriptions the description holder
	 * @param topic the topic to publish
	 */
	private void publishEndpointDescription(org.gecko.rsa.model.rsa.EndpointDescriptions descriptions, String topic) {
		try {
			Promise<OutputStream> payload = deSerializer.serialize(descriptions);
			ByteArrayOutputStream baos = (ByteArrayOutputStream) payload.getValue();
			messaging.publish(topic, ByteBuffer.wrap(baos.toByteArray()));
		} catch (Exception e) {
			logger.log(Level.SEVERE, String.format("Error publishing payload for endpoint to topic %s", topic), e);
		}
	}

	/**
	 * Publishes a single end-point description
	 * @param description the end-point description
	 * @param topic the topic to publish
	 */
	private void publishEndpointDescription(org.gecko.rsa.model.rsa.EndpointDescription description, String topic) {
		EndpointDescriptions descriptions = RSAFactory.eINSTANCE.createEndpointDescriptions();
		descriptions.getEndpointDescription().add(description);
		publishEndpointDescription(descriptions, topic);
	}

	/**
	 * Publishes all descriptions
	 * @param message the message with payload
	 */
	private void publishAllDescriptions(Message message) {
		String topicPart = new String(message.payload().array());
		if (frameworkUUID == null || frameworkUUID.equalsIgnoreCase(topicPart)) {
			logger.fine("Received a get all request for this OSGi framework (Requesting ourself). It can be ignored.");
			return;
		}
		String topic = String.format(MADiscoveryConstants.TOPIC_ANNOUNCED_SERVICE, topicPart);
		Collection<org.gecko.rsa.model.rsa.EndpointDescription> allDescriptions = exportingDescriptions.values();
		EndpointDescriptions descriptions = RSAFactory.eINSTANCE.createEndpointDescriptions();
		if (!allDescriptions.isEmpty()) {
			descriptions.getEndpointDescription().addAll(allDescriptions);
		}
		publishEndpointDescription(descriptions, topic);
	}
	
	/**
	 * Calls for new description
	 */
	private void callForDescription() {
		try {
			ByteBuffer buffer = ByteBuffer.wrap(frameworkUUID.getBytes());
			messaging.publish(MADiscoveryConstants.TOPIC_GET_DESCRIPTION, buffer);
		} catch (Exception e) {
			logger.log(Level.SEVERE, "Error requesting for service descriptions.", e);
		}
	}

}
